# -*- coding: utf-8 -*-
# Copyright (C) 2019 GIS OPS UG
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

__title__ = 'routingpy'
__description__ = 'One lib to route them all.'
__url__ = 'https://gis-ops.com/projects/routingpy'
__version__ = '0.2.7'
__author__ = 'Nils Nolde'
__author_email__ = 'nils@gis-ops.com'
__license__ = 'Apache 2.0'
__copyright__ = 'Copyright 2019 GIS OPS UG'
